#include "acctcl.h"

int Int64_Cmd (ClientData, Tcl_Interp *, int, char *[]) ;

int Int64_Add   (ClientData, Tcl_Interp *, int, char *[]) ;
int Int64_Sub   (ClientData, Tcl_Interp *, int, char *[]) ;
int Int64_Mul   (ClientData, Tcl_Interp *, int, char *[]) ;
int Int64_Div   (ClientData, Tcl_Interp *, int, char *[]) ;
int Int64_Mod   (ClientData, Tcl_Interp *, int, char *[]) ;
int Int64_Lt    (ClientData, Tcl_Interp *, int, char *[]) ;
int Int64_Le    (ClientData, Tcl_Interp *, int, char *[]) ;
int Int64_Gt    (ClientData, Tcl_Interp *, int, char *[]) ;
int Int64_Ge    (ClientData, Tcl_Interp *, int, char *[]) ;
int Int64_Eq    (ClientData, Tcl_Interp *, int, char *[]) ;
int Int64_Ne    (ClientData, Tcl_Interp *, int, char *[]) ;
int Int64_Print (ClientData, Tcl_Interp *, int, char *[]) ;

/******************************************************************************
Driver procedure
******************************************************************************/

int Int64_Cmd (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    static optable op [] =
    {
	{ "+",     Int64_Add   },
	{ "-",     Int64_Sub   },
	{ "*",     Int64_Mul   },
	{ "/",     Int64_Div   },
	{ "%",     Int64_Mod   },
	{ "<",     Int64_Lt    },
	{ ">",     Int64_Gt    },
	{ "<=",    Int64_Le    },
	{ ">=",    Int64_Ge    },
	{ "==",    Int64_Eq    },
	{ "!=",    Int64_Ne    },
	{ "add",   Int64_Add   },
	{ "sub",   Int64_Sub   },
	{ "mul",   Int64_Mul   },
	{ "div",   Int64_Div   },
	{ "mod",   Int64_Mod   },
	{ "print", Int64_Print },
    } ;

    return callproc (op, NTAB (op), cd, interp, argc, argv) ;
}

/******************************************************************************
Utility functions
******************************************************************************/

int print64 (Tcl_Interp *interp, char *str, int64 n, int base)
{
    char *start ;
    char *prefix ;
    int prefixlen ;

    switch (base)
    {
	case 8 :
	    prefix = "0"  ; prefixlen = 1 ; break ;
	case 10 :
	    prefix = ""   ; prefixlen = 0 ; break ;
	case 16 :
	    prefix = "0x" ; prefixlen = 2 ; break ;
	default :
	    Tcl_AppendResult(interp, "internal error: unexpected base conversion",
		(char *) NULL);
	    return TCL_ERROR ;
    }

    if (n < 0)
    {
	*str++ = '-' ;
	n = -n ;
    }
    strcpy (str, prefix) ;
    str = start = str + prefixlen ;

    do
    {
	int c ;

	c = n % base ;
	*str++ = c + (c  <= 9 ? '0' : 'a' - 10) ;
	n = n / base ;
    } while (n != 0) ;

    *str = '\0' ;
    str-- ;

    while (start < str)
    {
	int c ;

	c = *str ;
	*str = *start ;
	*start = c ;
	str-- ;
	start++ ;
    }

    return TCL_OK ;
}

int get64 (Tcl_Interp *interp, char *str, int64 *pn)
{
    int sign ;
    int base ;
    int64 n ;
    char *start ;

    sign = 1 ;
    n = 0 ;

    while (*str == '-' || *str == '+')
    {
	if (*str == '-')
	    sign = - sign ;
	str++ ;
    }

    if (str [0] == '0' && (str [1] == 'x' || str [1] == 'X'))
    {
	base = 16 ;
	str += 2 ;
    }
    else base = 10 ;

    start = str ;
    while (*str != '\0')
    {
	int c ;
	int cc ;

	c = *str ;
	cc = -1 ;

	if (c >= '0' && c <= '9')
	    cc = c - '0' ;
	else if (c >= 'a' && c <= 'f')
	    cc = (c - 'a') + 10 ;
	else if (c >= 'A' && c <= 'F')
	    cc = (c - 'A') + 10 ;

	if (cc == -1 || cc >= base)
	{
	    interp->result = "illegal digit" ;
	    return TCL_ERROR ;
	}

	n = n * base + cc ;
	str++ ;
    }

    if (start == str)
    {
	interp->result = "no digit found" ;
	return TCL_ERROR ;
    }

    n *= sign ;

    *pn = n ;
    return TCL_OK ;
}

int op64 (Tcl_Interp *interp, int argc, char *argv [], int op)
{
    int64 n1, n2, n3 ;
    char str [100] ;

    if (argc != 4)
    {
	interp->result = "wrong # args: should be int64 op n1 n2" ;
	return TCL_ERROR ;
    }

    if (get64 (interp, argv [2], &n1) == TCL_ERROR)
	return TCL_ERROR ;
    if (get64 (interp, argv [3], &n2) == TCL_ERROR)
	return TCL_ERROR ;
    
    switch (op)
    {
	case '+' : n3 = n1 + n2 ; break ;
	case '-' : n3 = n1 - n2 ; break ;
	case '*' : n3 = n1 * n2 ; break ;
	case '/' : n3 = n1 / n2 ; break ;
	case '%' : n3 = n1 % n2 ; break ;
	case '<' : n3 = n1 < n2 ; break ;
	case '>' : n3 = n1 > n2 ; break ;
	case 'l' : n3 = n1 <= n2 ; break ;
	case 'g' : n3 = n1 >= n2 ; break ;
	case '=' : n3 = n1 == n2 ; break ;
	case '#' : n3 = n1 != n2 ; break ;
    }

    if (print64 (interp, str, n3, 16) != TCL_OK)
	return TCL_ERROR ;
    Tcl_SetResult (interp, str, TCL_VOLATILE) ;
    return TCL_OK ;
}

/******************************************************************************
Command "int64 add/sub/mul/..."
******************************************************************************/

int Int64_Add (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    return op64 (interp, argc, argv, '+') ;
}

int Int64_Sub   (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    return op64 (interp, argc, argv, '-') ;
}

int Int64_Mul   (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    return op64 (interp, argc, argv, '*') ;
}

int Int64_Div   (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    return op64 (interp, argc, argv, '/') ;
}

int Int64_Mod   (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    return op64 (interp, argc, argv, '%') ;
}

int Int64_Lt    (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    return op64 (interp, argc, argv, '<') ;
}

int Int64_Le    (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    return op64 (interp, argc, argv, 'l') ;
}

int Int64_Gt    (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    return op64 (interp, argc, argv, '>') ;
}

int Int64_Ge    (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    return op64 (interp, argc, argv, 'g') ;
}

int Int64_Eq    (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    return op64 (interp, argc, argv, '=') ;
}

int Int64_Ne    (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    return op64 (interp, argc, argv, '#') ;
}


/******************************************************************************
Command "int64 print"
******************************************************************************/

int Int64_Print (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    char str [100] ;
    int64 n ;
    int i, r ;
    enum { HEX, DEC, ENG } mode ;
    int ndigits ;

    mode = DEC ;
    i = 2 ;
    while (argc >= 3 && argv [i][0] == '-')
    {
	if (strcmp (argv [i], "-hex") == 0)
	    mode = HEX ;
	else if (strcmp (argv [i], "-dec") == 0)
	    mode = DEC ;
	else if (strcmp (argv [i], "-eng") == 0)
	{
	    i++ ;
	    argc-- ;
	    if (argc < 2)
	    {
		interp->result =
			"wrong # args: option -eng needs a numeric argument" ;
		return TCL_ERROR ;
	    }
	    if (Tcl_GetBoolean (interp, argv [i], &ndigits) != TCL_OK)
		return TCL_ERROR ;
	    mode = ENG ;
	}
	else if (strcmp (argv [i], "--") == 0)
	{
		i++ ;
		argc-- ;
		break ;
	}
	else
	{
	    Tcl_ResetResult (interp) ;
	    Tcl_AppendResult (interp, "bad option \"", argv [i],
		": should be -hex, -dec or -eng",
		(char *) NULL) ;
	    return TCL_ERROR ;
	}
	i++ ;
	argc-- ;
    }

    if (argc != 3)
    {
	interp->result = "wrong # args: should be int64 print ?option? n" ;
	return TCL_ERROR ;
    }

    if (get64 (interp, argv [i], &n) == TCL_ERROR)
	return TCL_ERROR ;

    switch (mode)
    {
	case HEX :
	    r = print64 (interp, str, n, 16) ;
	    break ;
	case DEC :
	    r = print64 (interp, str, n, 10) ;
	    break ;
	case ENG :
	    /* **** r = print64eng (interp, str, n, ndig) ; */
	    r = TCL_ERROR ;
	    break ;
    }
    if (r != TCL_OK)
	return TCL_ERROR ;
    Tcl_SetResult (interp, str, TCL_VOLATILE) ;
    return TCL_OK ;
}
